/*
 * isobar.c: ISO Boot Archive Remover.
 *
 * Jason Hood, 6, 8 & 9 March, 2005.
 *
 * Extract the boot image (or code if no emulation) from a bootable CD-ROM
 * (or an image of one).
 *
 * Adapted from the program by David Brinkman.
 *
 * v1.01, 30 May, 2005:
 *   always include Image Size in display.
 *
 * v1.02, 5 & 6 June, 2005:
 *   Win32 port.
 *
 * v1.03, 10 Nov, 2007:
 *   Linux port by newhren <colimit@gmail.com>
 *
 * Todo: possibly replace the boot image in a CD image;
 *      recognise boot sections;
 *      ignore non-bootable CDs (are there any?).
 *
 * http://www.phoenix.com/NR/rdonlyres/98D3219C-9CC9-4DF5-B496-A286D893E36A/0/specscdrom.pdf
 * http://www.geocities.com/jadoxa/shsucdx/omibar32.zip
 * http://adoxa.110mb.com/shsucdx/omibar32.zip
 * http://colimit.googlepages.com/isobar.c
 *
 */

#define PVERS "1.03"
#define PDATE "10 Nov, 2007"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/stat.h>

typedef uint8_t  BYTE;
typedef uint16_t WORD;
typedef uint32_t DWORD;
typedef unsigned int  UINT;

#define MAX 32
#define OFLAG O_CREAT | O_WRONLY | O_TRUNC
#define OMODE S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH

char * buf;
DWORD offset = 0, imgsize;
int fdin;

int CDRead( UINT SectorCount, DWORD StartSector )
{
return (pread(fdin, buf, SectorCount << 11, StartSector << 11) == SectorCount << 11); // Sector size = 0x800
}

const char* platform[] = { "80x86",
			   "Power PC",
			   "Mac" };

const char* boot_type[] = { "no emulation",
			    "1.2 meg floppy",
			    "1.44 meg floppy",
			    "2.88 meg floppy",
			    "hard disk" };

enum
{
  E_OK, 		// No problems
  E_OPT,		// Unknown/invalid option
  E_MEM,		// Not enough memory
  E_NOCD,		// Not a CD drive, MSCDEX/SHSUCDX not installed,
			//  unknown CD format or no CD present
  E_CREATE,		// Unable to create image file
  E_ABORTED		// Read/write error
};


void usage( void )
{
  puts(

"ISOBAR, Version "PVERS" ("PDATE"). Freeware.\n"
"\n"
"Extract the boot image (or code) from .ISO image of a bootable CD-ROM.\n"
"\n"
"isobar [-o file] iso-file\n"
"\n"
"-o file       Write the boot image (or code) to the specified filename\n"
"                (without this, just boot information is displayed).\n"
"iso-file      An image of a bootable CD-ROM.\n"

  );

  exit( E_OK );
}


int main( int argc, char* argv[] )
{
  int	i;
  int	fdout = 0;
  char *outfile = NULL, *isofile = NULL;
  DWORD bootfound;
  int	type;
  DWORD n;
  UINT	len, w;


  if (argc > 1)
  {
    if (argv[1][0] == '?' || argv[1][1] == '?' || !strcmp( argv[1], "--help" ))
      usage();

    for (i = 1; i < argc; ++i)
    {
      if (*argv[i] == '-')
      {
	switch (argv[i][1] | 0x20)
	{
	  case 'o':
	    if (argv[i][2] == '\0' && argv[i+1] == NULL)
	    {
	      fputs( "ERROR: -o requires filename.\n", stderr );
	      return E_OPT;
	    }
	    outfile = (argv[i][2] == '\0') ? argv[++i] : argv[i] + 2;
	  break;

	  case 'h':
	    usage ();
	  break;

	  default:
	    fprintf( stderr, "ERROR: unknown option: %s.\n", argv[i] );
	    return E_OPT;
	}
      }
      else
      {
	isofile = argv[i];
      }
    }
  } else usage();


  fdin = open( isofile, O_RDONLY );
  
  if (fdin == -1)
  {
    fprintf( stderr, "ERROR: Cannot open %s.\n", isofile );
    return E_NOCD;
  }

  buf = malloc( MAX << 11 ); // transfer up to MAX Sectors (0x800) at a time
  if (buf == NULL)
  {
    fputs( "ERROR: Not enough memory.\n", stderr );
    return E_MEM;
  }

  if (!CDRead( 1, 0x11 )) // read the boot record
  {
    fputs( "Read error!\n", stderr );
    return E_ABORTED;
  }

  if (strcmp( buf+1, "CD001\01EL TORITO SPECIFICATION" ))
  {
    fprintf( stderr, "ERROR: %s is not EL TORITO.\n", isofile );
    return E_NOCD;
  }
  bootfound = *(DWORD *)(buf+0x47);
  printf( "Catalog Sector:\t%lx\n", (long unsigned int)bootfound );

  if (!CDRead( 1, bootfound )) // read the boot catalog
  {
    fputs( "Read error!\n", stderr );
    return E_ABORTED;
  }

  // Just check the key bytes, don't worry about the checksum.
  if (buf[0x1e] != (char)0x55 || buf[0x1f] != (char)0xAA)
  {
    fprintf( stderr, "ERROR: %s has an invalid boot catalog.\n", isofile );
    return E_NOCD;
  }
  type = (BYTE)buf[1];
  printf( "Platform:\t%s (%02x)\n", (type < 3) ? platform[type] : "unknown",
				    type );
  printf( "ID String:\t%.24s\n", (buf[4]) ? buf+4 : "not recorded" );
  printf( "Bootable:\t%s (%02x)\n", (buf[0x20] == (char)0x88) ? "yes" : "no",
				    (BYTE)buf[0x20] );
  type = buf[0x21] & 15;
  printf( "Boot Type:\t%s (%02x)\n", (type < 5) ? boot_type[type] : "unknown",
				     (BYTE)buf[0x21] );
  i = *(WORD *)(buf+0x22);
  printf( "Load Segment:\t%04x\n", (i == 0) ? 0x7c0 : i );
  printf( "System Type:\t%02x\n", (BYTE)buf[0x24] );
  imgsize = *(WORD *)(buf+0x26);
  printf( "Sector Count:\t%02x (%d)\n", (UINT)imgsize, (UINT)imgsize );
  offset = *(DWORD *)(buf+0x28);
  printf( "Image Sector:\t%lx\n", (long unsigned int)offset );

  switch (type)
  {
    case 0: // no emulation
    break;

    case 1: // 1.2meg floppy
	imgsize=1200*2;
    break;

    case 2: // 1.44meg floppy
	imgsize=1440*2;
    break;

    case 3: // 2.88meg floppy
	imgsize=2880*2;
    break;

    case 4: // hard disk
	CDRead( 1, offset ); // read the MBR
	imgsize = *(DWORD *)(buf+0x1c6) + *(DWORD *)(buf+0x1ca); // offset + size of the first partition
    break;

    default:
	fprintf( stderr, "ERROR: EL TORITO boot media type not recognized\n", isofile );
	return E_NOCD;
  }

  imgsize *= 0x200; // Virtual Sector size = 0x200
  printf( "Image Size:\t%ld bytes\n", (long int)imgsize );

  if (outfile)
  {
    fdout = open( outfile, OFLAG, OMODE );
    if (fdout < 0)
    {
      fprintf( stderr, "ERROR: Cannot create %s.\n", outfile );
      return E_CREATE;
    }

    do
    {
      n = imgsize >> 11; // Sector size = 0x800
      if (n > MAX)
	n = MAX;
      else if (n == 0)
	n = 1;
      if (!CDRead( (UINT)n, offset ))
      {
	fputs( "Read error!", stderr );
	return E_ABORTED;
      }
      len = (UINT)n << 11;
      if (len > imgsize)
	len = (UINT)imgsize;
      w = write( fdout, buf, len );
      if (w != len)
      {
	fputs( "Write error!", stderr );
	return E_ABORTED;
      }
      offset += n;
      imgsize -= len;
    } while (imgsize);
    close( fdout );
    printf( "\nThe output image has been saved in: %s\n", outfile );
  }

  close( fdin );
  return 0;
}
